// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file TFQMR.h */

#ifndef AMDIS_TFQMR_H
#define AMDIS_TFQMR_H

#include "OEMSolver.h"
#include "MemoryManager.h"

namespace AMDiS {

  // ============================================================================
  // ===== class TFQMR ==========================================================
  // ============================================================================

  /** 
   * \ingroup Solver
   *
   * \brief
   */
  template<typename VectorType>
  class TFQMR : public OEMSolver<VectorType>
  {
  public:
    MEMORY_MANAGED(GMResSolver2<VectorType>);

    /** \brief
     * Creator class used in the OEMSolverMap.
     */
    class Creator : public OEMSolverCreator<VectorType>
    {
    public:
      MEMORY_MANAGED(Creator);

      virtual ~Creator() {};

      /** \brief
       * Returns a new TFQMR object.
       */   
      OEMSolver<VectorType>* create() { 
	return NEW TFQMR<VectorType>(this->name);
      };
    };

    /** \brief
     * constructor
     */
    TFQMR(std::string name);

    /** \brief
     * destructor
     */
    ~TFQMR();

  protected:
    /** \brief
     * realisation of OEMSolver::solveSystem
     */
    int solveSystem(MatVecMultiplier<VectorType> *mv, 
		    VectorType *x, VectorType *b, bool reuseMatrix);

    /** \brief
     * realisation of OEMSolver::init
     */
    void init();

    /** \brief
     * realisation of OEMSolver::exit
     */
    void exit();

  private:
    /** \brief
     * Stores the tolerance boundary for numerical computations.
     */
    double TOL_;

    VectorType *r0_;

    /** \brief
     * Stores intermediate results in the orthogonalization step.
     */
    VectorType *w_;

    /** \brief
     * Pointers to the vectors of the orthogonal system.
     */
    VectorType *y1_, *y2_;

    VectorType *v_, *d_, *tmp_;

  };


}

#include "TFQMR.hh"

#endif // AMDIS_TFQMR_H
